# -*- coding: utf-8 -*-

require 'sidfm_controller_support'

class SidfmCtagsController < ApplicationController
  unloadable

  layout 'sidfm'

  self.main_menu = false

  protect_from_forgery with: :exception

  helper :sidfm

  private

  include SidfmControllerSupport

  before_action :update_session_updated_at
  before_action :get_sidfm_roles
  before_action :create_common_data

  before_action :check_license
  before_action :check_roles

  before_action :get_vm_version
  before_action :clear_import_session
  before_action :clear_export_session

  before_action -> { get_summary_table_lines('ctags') }

  # 権限チェック
  def check_roles
    unless @sidfm_user then
      redirect_to home_path, :notice => af_log_user('許可されていない操作です。') and return
    end
    unless @sidfm_roles.include?(:view) then
      redirect_to user_path(User.current), :notice => af_log_user('アクセス権限がありません。') and return
    end
    unless @sidfm_roles.include?(:host_admin) then
      if [ 'create', 'modify', 'delete', 'bulk_delete' ].include?(action_name) then
        redirect_to sidfm_ctags_path, :notice => af_log_user('許可されていない操作です。') and return
      end
    end
    @role_admin = @sidfm_roles.include?(:admin)
    @user_category_ids = if @role_admin then
                           nil
                         else
                           SidfmdCategoriesUser.where(:user_id => User.current.id)
                                               .select(:sidfmd_category_id)
                                               .order(:sidfmd_category_id)
                                               .distinct
                                               .pluck(:sidfmd_category_id)
                         end
    @user_host_ids = if @role_admin then
                       nil
                     else
                       SidfmdCategoriesHost.joins(:sidfmd_host)
                                           .where(:sidfmd_category_id => @user_category_ids)
                                           .where(:sidfmd_hosts => { :flg_delete => false })
                                           .select(:sidfmd_host_id)
                                           .order(:sidfmd_host_id)
                                           .distinct
                                           .pluck(:sidfmd_host_id)
                     end
  end

  def parse_params
    f_params = params.fetch(:pages, {}).permit(:order, :offset, :per_page, :filters, :checked, :all_checked)
    @ctags = {}
    @ctags[:per_page] = (@summary_table_lines || f_params[:per_page] || 10).to_i
    @ctags[:offset] = (f_params[:offset] || 0).to_i
    @ctags[:order] = f_params[:order] || 'priority_asc'
    @ctags[:filters] = JSON.parse((f_params[:filters] || '{}'), symbolize_names: true)
    @ctags[:checked] = JSON.parse(f_params[:checked] || '[]')
    @ctags[:all_checked] = f_params[:all_checked] || false
  end

  # ソート
  def sort_content
    last_index = @ctags[:order].rindex('_')
    col = @ctags[:order].slice(0..last_index - 1)
    order = @ctags[:order].slice(last_index + 1..@ctags[:order].size - 1)
    
    case col
    when 'priority' then
      @ctags[:all_content] = @ctags[:all_content].order(:priority => (order == 'asc' ? 'desc' : 'asc').to_sym)
                                                 .order(%|
        LOWER(keyword COLLATE "C") #{order.upcase}
      |)
    when 'name' then
      @ctags[:all_content] = @ctags[:all_content].order(%|
        LOWER(keyword COLLATE "C") #{order.upcase}
      |)
      .order(:priority => (order == 'asc' ? 'desc' : 'asc').to_sym)
    when 'category' then
      @ctags[:all_content] = @ctags[:all_content].order(%|
        LOWER(sidfmd_categories.name COLLATE "C") #{order.upcase}
      |)
      .order(:priority => order.to_sym)
      .order(%|
        LOWER(keyword COLLATE "C") #{order.upcase}
      |)
    end
    
  end

  FILTER_KEYS = {}

  def init_filter_keys
    if FILTER_KEYS.size == 0 then
      FILTER_KEYS[:name] = ''
      FILTER_KEYS[:category_id] = '0'
    end
  end

  def check_filters
    init_filter_keys
    FILTER_KEYS.each do |key, val|
      unless @ctags[:filters].has_key?(key) then
        @ctags[:filters][key] = val
      end
    end
    @default_filters = FILTER_KEYS
  end

  # フィルター
  def filter_content
    if @ctags[:filters][:name].length > 0 then
      @ctags[:all_content] = @ctags[:all_content].where("LOWER(keyword) LIKE ?", '%' + ActiveRecord::Base.send(:sanitize_sql_like, @ctags[:filters][:name].downcase) + '%')
    end

    if @ctags[:filters][:category_id].to_i != 0 then
      @ctags[:all_content] = @ctags[:filters][:category_id].to_i == -1 ?
        @ctags[:all_content].where("sidfmd_categories.id IS NULL") :
        @ctags[:all_content].where("sidfmd_categories.id = ?", @ctags[:filters][:category_id].to_i)
    end
  end
 
  def create_content
    @ctags[:all_content] = SidfmdCtag.create_content(
      :role_admin => @role_admin,
      :user_category_ids => @user_category_ids
    )
  end

  def create_summary_ctag_html(ctag, **opts)
    data = {
      :chkbox => view_context.create_summary_ctag_chkbox(ctag.id, opts[:checked], (!@sidfm_roles.include?(:host_admin) || (!@role_admin && !ctag.category_id))),
      :priority => view_context.create_summary_ctag_priority(ctag.priority),
      :name => view_context.create_summary_ctag_name(ctag.name),
      :category => view_context.create_summary_ctag_category(ctag.category_name, ctag.category_desc),
      :color => view_context.create_summary_ctag_color(ctag.color, ctag.bgcolor),
      :edit => view_context.create_summary_ctag_edit(ctag, (!@sidfm_roles.include?(:host_admin) || (!@role_admin && !ctag.category_id)))
    }
  end

  def is_valid_color_code(color_code)
    !!(/^#([A-Fa-f0-9]{6}|[A-Fa-f0-9]{3})$/ =~ color_code)
  end

  def get_regist_data(filtered_params)
    cat = (filtered_params[:category_id] && filtered_params[:category_id].length > 0) ?
      SidfmdCategory.find_by(:id => filtered_params[:category_id]) :
      nil
    cat_id = cat ? cat.id : nil
    data = {
      :sidfmd_category_id => cat_id,
      :keyword => filtered_params[:name],
      :priority => filtered_params[:priority].to_i,
      :color => (filtered_params[:text_color] ? filtered_params[:text_color] : ''),
      :bgcolor => (filtered_params[:bg_color] ? filtered_params[:bg_color] : '')
    }
  end

  # タグのカテゴリ更新時・削除にViewを更新
  def update_issue_related_views(sidfmd_issue_ids)
    update_view_data = SidfmdIssue.where(:id => sidfmd_issue_ids).map{|issue|
      {
        :sidfmd_issue_id => issue.id,
        :html_id => issue.html_id,
        :sidfmd_host_id => issue.sidfmd_host_id,
      }
    }
    SidfmdVulView.update_from_rule_action(update_view_data)
    SidfmdHostView.update_from_rule_action(update_view_data)
  end

  # 関連するルールを処理
  def process_related_rules(**opts)
    r = true
    param = {
      :role_admin => @role_admin,
      :user_category_ids => (@role_admin ? nil : @user_category_ids),
      :action => opts[:action],
      :messages_necessary => false
    }
    case opts[:action]
    when 'delete' then
      param[:delete_ctag_ids] = opts[:delete_ctag_ids]
    when 'category_change' then
      param[:change_category_ctag_ids] = opts[:change_category_ctag_ids]
    end
    ret = SidfmdRule.ctags_related_rule_info(param)
    if ret[:is_unchangeable_ctag] then
      r = false
    else
      SidfmdRuleCmCtag.where(:id => ret[:delete_cm_ctag_ids]).destroy_all if ret[:delete_cm_ctag_ids].any?
      SidfmdRuleAmCtag.where(:id => ret[:delete_am_ctag_ids]).destroy_all if ret[:delete_am_ctag_ids].any?
      SidfmdRuleCond.where(:id => ret[:delete_cond_ids]).destroy_all if ret[:delete_cond_ids].any?
      SidfmdRuleAction.where(:id => ret[:delete_action_ids]).destroy_all if ret[:delete_action_ids].any?
      SidfmdRule.where(:id => ret[:delete_rule_ids]).destroy_all if ret[:delete_rule_ids].any?
    end
    r
  end

  public

  def index
    parse_params
    check_filters
    @page_name = 'カスタムタグ設定'
  end

  def create
    f_params = params.permit(:name, :priority, :category_id, :text_color, :bg_color)
    if !f_params[:name] or f_params[:name].length == 0 or !f_params[:priority] or !f_params[:priority].length then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    begin
      data = get_regist_data(f_params)
      if (data[:color].length > 0 && !is_valid_color_code(data[:color])) ||
         (data[:bgcolor].length > 0 && !is_valid_color_code(data[:bgcolor])) ||
         (!@role_admin && !data[:sidfmd_category_id])
      then
        render :layout => false, :json => { :err_message => '不正な操作です。' } and return
      end
      if !@role_admin && !@user_category_ids.include?(f_params[:category_id].to_i) then
        render :layout => false, :json => { :err_message => '不正な操作です。' } and return
      end
      is_same_keyword = SidfmdCtag.is_same_keyword(f_params[:name])
      if is_same_keyword then
        render :layout => false, :json => { :err_message => '既に登録済みのタグ名です。' } and return
      end
      SidfmdCtag.transaction do
        SidfmdCtag.create(data)
      end
    rescue Exception => e
      log_exception(e)
      render :layout => false, :json => { :err_message => 'タグの作成に失敗しました。' } and return
    end

    render :layout => false, :json => { :err_message => '' }
  end

  def modify
    f_params = params.permit(:id, :name, :priority, :category_id, :text_color, :bg_color)
    if !f_params[:id] or !f_params[:name] or f_params[:name].length == 0 or !f_params[:priority] or !f_params[:priority].length then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end

    deleted_sidfmd_issue_ids = []
    begin
      data = get_regist_data(f_params)
      if (data[:color].length > 0 && !is_valid_color_code(data[:color])) ||
         (data[:bgcolor].length > 0 && !is_valid_color_code(data[:bgcolor])) ||
         (!@role_admin && !data[:sidfmd_category_id])
      then
        render :layout => false, :json => { :err_message => '不正な操作です。' } and return
      end
      unless @role_admin then
        user_ctag_ids = SidfmdCtag.user_ctag_ids(@role_admin, @user_category_ids)
        if !user_ctag_ids.include?(f_params[:id].to_i) then
          render :layout => false, :json => { :err_message => '不正な操作です。' } and return
        end
        if !@user_category_ids.include?(f_params[:category_id].to_i) then
          render :layout => false, :json => { :err_message => '不正な操作です。' } and return
        end
      end
      is_same_keyword = SidfmdCtag.is_same_keyword(f_params[:name], f_params[:id].to_i)
      if is_same_keyword then
        render :layout => false, :json => { :err_message => '既に登録済みのタグ名です。' } and return
      end
      SidfmdCtag.transaction do
        ctag = SidfmdCtag.find_by(:id => f_params[:id].to_i)
        if ctag then
          prev_category_id = ctag.sidfmd_category_id
          new_category_id = data[:sidfmd_category_id]
          ctag.update(data)
          if prev_category_id != new_category_id then
            deleted_sidfmd_issue_ids = ctag.delete_ctag_vul_related_rows
            # 関連するルールの処理
            process_rules = process_related_rules(
              :change_category_ctag_ids => { f_params[:id].to_i => new_category_id },
              :action => 'category_change'
            )
            if !process_rules then
              logger.error('failed to process rules')
              raise Exception
            end
          end
        else
          render :layout => false, :json => { :err_message => 'タグが存在しません。' } and return
        end
      end
    rescue Exception => e
      log_exception(e)
      render :layout => false, :json => { :err_message => 'タグの更新に失敗しました。' } and return
    end

    if deleted_sidfmd_issue_ids.any? then
      begin
        update_issue_related_views(deleted_sidfmd_issue_ids)
      rescue Exception => e
        log_exception(e)
      end
    end

    render :layout => false, :json => { :err_message => '' }
  end

  def delete
    f_params = params.permit(:id)
    if !f_params[:id] then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    unless @role_admin then
      user_ctag_ids = SidfmdCtag.user_ctag_ids(@role_admin, @user_category_ids)
      if !user_ctag_ids.include?(f_params[:id].to_i) then
        render :layout => false, :json => { :err_message => '不正な操作です。' } and return
      end
    end

    deleted_sidfmd_issue_ids = []
    begin
      SidfmdCtag.transaction do
        ctag = SidfmdCtag.find_by(:id => f_params[:id].to_i)
        if ctag then
          #事前にsidfmd_issue_ctags, sidfmd_vv_ctags, sidfmd_hv_ctags及び関連テーブルの対応するレコードを事前に削除する
          deleted_sidfmd_issue_ids = ctag.delete_ctag_vul_related_rows
          process_rules = process_related_rules(
            :delete_ctag_ids => [ f_params[:id].to_i ],
            :action => 'delete'
          )
          if !process_rules then
            logger.error('failed to process rules')
            raise Exception
          end
          ctag.destroy
        end
      end
    rescue Exception => e
      log_exception(e)
      render :layout => false, :json => { :err_message => '削除に失敗しました。' } and return
    end

    if deleted_sidfmd_issue_ids.any? then
      begin
        update_issue_related_views(deleted_sidfmd_issue_ids)
      rescue Exception => e
        log_exception(e)
      end
    end

    render :layout => false, :json => { :err_message => '' }
  end

  def bulk_delete
    f_params = params.permit(:ids)
    if !f_params[:ids] then
      render :layout => false, :json => { :err_message => '不正な操作です。' } and return
    end
    begin
      ids = JSON.parse(f_params[:ids])
      ids = ids.map(&:to_i)
      unless @role_admin then
        user_ctag_ids = SidfmdCtag.user_ctag_ids(@role_admin, @user_category_ids)
        if (ids - user_ctag_ids).any? then
          render :layout => false, :json => { :err_message => '不正な操作です。' } and return
        end
      end
      SidfmdCtag.transaction do
        #事前にsidfmd_issue_ctags, sidfmd_vv_ctags, sidfmd_hv_ctags及び関連テーブルの対応するレコードを事前に削除する
        process_rules = process_related_rules(
          :delete_ctag_ids => ids,
          :action => 'delete'
        )
        if !process_rules then
          logger.error('failed to process rules')
          raise Exception
        end
        SidfmdCtag.delete_ctags(ids)
      end
    rescue Exception => e
      log_exception(e)
      render :layout => false, :json => { :err_message => '削除に失敗しました。' } and return
    end
    
    begin
      update_issue_related_views(ids)
    rescue Exception => e
      log_exception(e)
    end

    render :layout => false, :json => { :err_message => '' }
  end

  def api_get_display_summary_content
    GC.start
    f_params = params.fetch(:pages, {}).permit(:per_page)
    parse_params
    @ctags[:per_page] = f_params[:per_page].to_i
    create_content
    check_filters
    sort_content
    filter_content
    @ctags[:content_filtered] = @ctags[:all_content]
    @ctags[:content] = @ctags[:content_filtered].limit(@ctags[:per_page]).offset(@ctags[:offset])
    all_ids = [] #サマリー表で操作可能なid
    all_names = {}
    total = 0
    @ctags[:content_filtered].each do |ctag|
      if @role_admin || (@sidfm_roles.include?(:host_admin) && ctag.category_id) then
        all_ids << ctag.id
      end
      total += 1
      all_names[ctag.id] = ctag.name
    end

    ret = {
      :rows => @ctags[:content].map{|ctag|
        create_summary_ctag_html(
          ctag,
          {
            :checked => @ctags[:checked]
          }
        )
      },
      :total => total,
      :all_ids => all_ids,
      :all_names => all_names,
      :sidfm_roles => @sidfm_roles,
      :user_cats => @user_category_ids
    }
    render :layout => false, :json => ret
  end

  #同じタグ名(keyword)がないかチェック
  def api_is_same_keyword
    f_params = params.permit(:keyword, :modify_id)
    is_same_keyword = SidfmdCtag.is_same_keyword(f_params[:keyword], f_params[:modify_id])
    render :layout => false, :json => { :is_same_keyword => is_same_keyword }
  end

  #脆弱性と紐づいているかチェック
  def api_issue_related_ctags
    f_params = params.permit(:ids)
    ctag_ids = JSON.parse(f_params[:ids])
    ctags = SidfmdCtag.issue_related_ctags(ctag_ids)
    render :layout => false, :json => ctags
  end

end
